package org.processmining.glueminer.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.glueminer.models.Behavior;
import org.processmining.glueminer.models.Behaviors;
import org.processmining.glueminer.models.PlaceUnderfednessStorage;
import org.processmining.glueminer.models.PlaceUnderfednessStorageComparator;
import org.processmining.glueminer.models.PlaceUsageStatus;
import org.processmining.glueminer.models.PlaceUsageStatus.status;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;

public class TokenReplayer {
	public Behaviors behaviors;
	public Petrinet net;
	public Marking marking;
	public Marking finalMarking;
	public Map<Transition, List<Place>> inputPlacesForTransition;
	public Map<Transition, List<Place>> outputPlacesForTransition;
	public Map<Place, Set<Behavior>> fitBehaviorsForPlaces;
	public Map<Place, Set<Behavior>> underfedBehaviorsForPlaces;
	public Map<Place, Set<Behavior>> overfedBehaviorsForPlaces;
	public Map<Place, Integer> placeSupport;
	public Map<Place, Double> placeFitness;
	public Map<Place, Double> placeUnderfedness;
	public Map<Place, PlaceUsageStatus.status> placeUsageStatus;
	
	public Integer overallTraces;
	public Integer fitTraces;
	public double traceFitness;
	
	public TokenReplayer(Behaviors behaviors, Petrinet net, Marking initialMarking, Marking finalMarking) {
		this.behaviors = behaviors;
		this.net = net;
		this.marking = new Marking(initialMarking);
		this.finalMarking = finalMarking;
		this.reset();
	}
	
	public void reset() {
		this.fitBehaviorsForPlaces = new HashMap<Place, Set<Behavior>>();
		this.underfedBehaviorsForPlaces = new HashMap<Place, Set<Behavior>>();
		this.overfedBehaviorsForPlaces = new HashMap<Place, Set<Behavior>>();
		
		this.placeSupport = new HashMap<Place, Integer>();
		this.placeFitness = new HashMap<Place, Double>();
		this.placeUnderfedness = new HashMap<Place, Double>();
		this.placeUsageStatus = new HashMap<Place, PlaceUsageStatus.status>();
		
		this.inputPlacesForTransition = new HashMap<Transition, List<Place>>();
		this.outputPlacesForTransition = new HashMap<Transition, List<Place>>();
		
		this.overallTraces = 0;
		this.fitTraces = 0;
		this.traceFitness = 0.0;
		
		this.populatePlacesMaps();
		this.populateTransitionsMaps();
	}
	
	public void populatePlacesMaps() {
		for (Place p : net.getPlaces()) {
			this.fitBehaviorsForPlaces.put(p, new HashSet<Behavior>());
			this.underfedBehaviorsForPlaces.put(p, new HashSet<Behavior>());
			this.overfedBehaviorsForPlaces.put(p, new HashSet<Behavior>());
			this.placeSupport.put(p, 0);
			this.placeFitness.put(p, 0.0);
			this.placeUnderfedness.put(p, 0.0);
			this.placeUsageStatus.put(p, PlaceUsageStatus.status.NOTKNOWN);
		}
	}
	
	public void populateTransitionsMaps() {
		for (Transition t : net.getTransitions()) {
			this.inputPlacesForTransition.put(t, new ArrayList<Place>());
			this.outputPlacesForTransition.put(t, new ArrayList<Place>());
			
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inputEdges = net.getInEdges(t);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outputEdges = net.getOutEdges(t);
			
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inputEdge : inputEdges) {
				Place p = (Place) inputEdge.getSource();
				this.inputPlacesForTransition.get(t).add(p);
			}
			
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> outputEdge : outputEdges) {
				Place p = (Place) outputEdge.getTarget();
				this.outputPlacesForTransition.get(t).add(p);
			}
		}
	}
	
	public void doReplay() {
		for (Behavior b : behaviors.getBehaviorList()) {
			Marking thisMarking = new Marking(marking);
			Set<Place> allInvolvedPlaces = new HashSet<Place>();
			Set<Place> consumptionProblems = new HashSet<Place>();
			
			for (Transition t : b.enabledTransitions) {
				List<Place> sourcePlaces = this.inputPlacesForTransition.get(t);
				List<Place> targetPlaces = this.outputPlacesForTransition.get(t);
				
				for (Place sourcePlace : sourcePlaces) {
					allInvolvedPlaces.add(sourcePlace);
					if (thisMarking.contains(sourcePlace)) {
						thisMarking.remove(sourcePlace);
					}
					else {
						consumptionProblems.add(sourcePlace);
					}
				}
				
				for (Place targetPlace : targetPlaces) {
					allInvolvedPlaces.add(targetPlace);
					thisMarking.add(targetPlace);
				}
			}
			
			thisMarking.removeAll(finalMarking);
			
			for (Place place : allInvolvedPlaces) {
				if (consumptionProblems.contains(place)) {
					this.underfedBehaviorsForPlaces.get(place).add(b);
				}
				else if (thisMarking.contains(place)) {
					this.overfedBehaviorsForPlaces.get(place).add(b);
				}
				else {
					this.fitBehaviorsForPlaces.get(place).add(b);
				}
			}
			
			overallTraces += b.getVariantCount();
			
			if (consumptionProblems.size() == 0) {
				fitTraces += b.getVariantCount();
			}
			
			/*if (thisMarking.equals(finalMarking)) {
				for (Place place : consumptionProblems) {
					
				}
			}
			else {
				for (Place place : consumptionProblems) {
					
				}
			}*/
		}
		calculateStatistics();
	}

	public int getSumOfBehaviorsCount(Set<Behavior> behaviors) {
		int ret = 0;
		for (Behavior b : behaviors) {
			ret += b.getVariantCount();
		}
		return ret;
	}
	
	public void calculateStatistics() {
		if (this.overallTraces > 0) {
			this.traceFitness = (double)this.fitTraces / (double)this.overallTraces;
		}
		for (Place p : fitBehaviorsForPlaces.keySet()) {
			int fit = getSumOfBehaviorsCount(this.fitBehaviorsForPlaces.get(p));
			int underfed = getSumOfBehaviorsCount(this.underfedBehaviorsForPlaces.get(p));
			int overfed = getSumOfBehaviorsCount(this.overfedBehaviorsForPlaces.get(p));
			int sum = fit + underfed + overfed;	
			double placeFitness = 0.0;
			double placeUnderfedness = 0.0;
			if (sum > 0) {
				placeFitness = (double)fit / (double)sum;
				placeUnderfedness = (double)underfed / (double)sum;
			}
			this.placeFitness.put(p, placeFitness);
			this.placeUnderfedness.put(p, placeUnderfedness);
			this.placeSupport.put(p, fit+underfed+overfed);
			if (underfed >= overfed && underfed >= fit) {
				this.placeUsageStatus.put(p, PlaceUsageStatus.status.UNDERFED);
			}
			else if (overfed >= underfed && overfed >= fit) {
				this.placeUsageStatus.put(p, PlaceUsageStatus.status.OVERFED);
			}
			else {
				this.placeUsageStatus.put(p, PlaceUsageStatus.status.FIT);
			}
		}
	}
	
	public List<PlaceUnderfednessStorage> getPlaceUnderfednessStorage() {
		List<PlaceUnderfednessStorage> storage = new ArrayList<PlaceUnderfednessStorage>();
		for (Place p : this.placeUnderfedness.keySet()) {
			if (!p.getLabel().equals("source") && !p.getLabel().equals("sink")) {
				storage.add(new PlaceUnderfednessStorage(p, this.placeUnderfedness.get(p)));
			}
		}
		Collections.sort(storage, new PlaceUnderfednessStorageComparator());
		
		return storage;
	}
	
	public void printResume() {
		for (Place p : this.fitBehaviorsForPlaces.keySet()) {
			System.out.println("\n\nPlace "+p.getLabel());
			System.out.println("Fit behaviors: "+getSumOfBehaviorsCount(this.fitBehaviorsForPlaces.get(p)));
			System.out.println("Underfed behaviors: "+getSumOfBehaviorsCount(this.underfedBehaviorsForPlaces.get(p)));
			System.out.println("Overfed behaviors: "+getSumOfBehaviorsCount(this.overfedBehaviorsForPlaces.get(p)));
			System.out.println("Overall behavior: "+this.placeUsageStatus.get(p));
			System.out.println("Place fitness: "+this.placeFitness.get(p));
			System.out.println("Place underfedness: "+this.placeUnderfedness.get(p));
		}
		System.out.println("\nTraces fitness: "+this.traceFitness+"\n");
	}
}
